/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "TComBaseLayerFilter.h"

#include <string.h>
#include <cassert>
#include <vector>
#include <math.h>

using namespace std;

#define BLF_DEF_PATTERN_IDX      5
#define BLF_COEF_ACC_BITS        8
#define BLF_COEF_ACC             (1<<BLF_COEF_ACC_BITS)

BLFParam TComBaseLayerFilter::m_blfParam;

static const BLFPattern blfPattern[] =
{
  { // (0) 7x7 square -- plain
    7, 49, 3, 3,
    {
      {  0, 1, 2, 3, 4, 5, 6, }, // ooooooo
      {  7, 8, 9,10,11,12,13, }, // ooooooo
      { 14,15,16,17,18,19,20, }, // ooooooo
      { 21,22,23,48,24,25,26, }, // oooxooo
      { 27,28,29,30,31,32,33, }, // ooooooo
      { 34,35,36,37,38,39,40, }, // ooooooo
      { 41,42,43,44,45,46,47, }, // ooooooo
    },
  },
  { // (1) 7x7 square -- point symmetry
    7, 25, 3, 3,
    {
      {  0, 1, 2, 3, 4, 5, 6, }, // ooooooo
      {  7, 8, 9,10,11,12,13, }, // ooooooo
      { 14,15,16,17,18,19,20, }, // ooooooo
      { 21,22,23,24,23,22,21, }, // ooox+++
      { 20,19,18,17,16,15,14, }, // +++++++
      { 13,12,11,10, 9, 8, 7, }, // +++++++
      {  6, 5, 4, 3, 2, 1, 0, }, // +++++++
    },
  },
  { // (2) 7x7 square -- x- and y-axis symmetry
    7, 16, 3, 3,
    {
      {  0, 1, 2, 3, 2, 1, 0, }, // oooo+++
      {  4, 5, 6, 7, 6, 5, 4, }, // oooo+++
      {  8, 9,10,11,10, 9, 8, }, // oooo+++
      { 12,13,14,15,14,13,12, }, // ooox+++
      {  8, 9,10,11,10, 9, 8, }, // +++++++
      {  4, 5, 6, 7, 6, 5, 4, }, // +++++++
      {  0, 1, 2, 3, 2, 1, 0, }, // +++++++
    },
  },
  { // (3) 7x7 square 
    7, 21, 3, 3,
    {
      {  0, 1, 2, 3, 2, 1, 0, }, // oooo+++
      {  4, 5, 6, 7, 6, 5, 4, }, // oooo+++
      {  8, 9,10,11,12, 9, 8, }, // ooooo++
      { 13,14,15,20,16,14,13, }, // oooxo++
      {  8, 9,17,18,19, 9, 8, }, // ++ooo++
      {  4, 5, 6, 7, 6, 5, 4, }, // +++++++
      {  0, 1, 2, 3, 2, 1, 0, }, // +++++++
    },
  },
  { // (4) 5x5 square -- plain
    5, 25, 2, 2,
    {
      {  0, 1, 2, 3, 4, }, // ooooo
      {  5, 6, 7, 8, 9, }, // ooooo
      { 10,11,24,12,13, }, // ooxoo
      { 14,15,16,17,18, }, // ooooo
      { 19,20,21,22,23, }, // ooooo
    },
  },
  { // (5) 5x5 square -- point symmetry
    5, 13, 2, 2,
    {
      {  0, 1, 2, 3, 4, }, // ooooo
      {  5, 6, 7, 8, 9, }, // ooooo
      { 10,11,12,11,10, }, // oox++
      {  9, 8, 7, 6, 5, }, // +++++
      {  4, 3, 2, 1, 0, }, // +++++
    },
  },
  { // (6) 5x5 square -- x- and y-axis symmetry
    5, 9, 2, 2,
    {
      {  0, 1, 2, 1, 0, }, // ooo++
      {  3, 4, 5, 4, 3, }, // ooo++
      {  6, 7, 8, 7, 6, }, // oox++
      {  3, 4, 5, 4, 3, }, // +++++
      {  0, 1, 2, 1, 0, }, // +++++
    },
  },
  { // (7) 7x7 star
    7, 13, 3, 3,
    {
      {  0,-1,-1, 1,-1,-1, 2, }, // o..o..o
      { -1, 3,-1, 4,-1, 5,-1, }, // .o.o.o.
      { -1,-1, 6, 7, 8,-1,-1, }, // ..ooo..
      {  9,10,11,12,11,10, 9, }, // ooox+++
      { -1,-1, 8, 7, 6,-1,-1, }, // ..+++..
      { -1, 5,-1, 4,-1, 3,-1, }, // .+.+.+.
      {  2,-1,-1, 1,-1,-1, 0, }, // +..+..+
    },
  },
  { // (8) 5x5 star
    5, 9, 2, 2,
    {
      {  0,-1, 1,-1, 2, }, // o.o.o
      { -1, 3, 4, 5,-1, }, // .ooo.
      {  6, 7, 8, 7, 6, }, // oox++
      { -1, 5, 4, 3,-1, }, // .+++.
      {  2,-1, 1,-1, 0, }, // +.+.+
    },
  },
  { // (9) 7x7cross + 3x3 square
    7, 17, 3, 3,
    {
      { -1,-1,-1, 0,-1,-1,-1, }, // ...o...
      { -1,-1,-1, 1,-1,-1,-1, }, // ...o...
      { -1,-1, 2, 3, 4,-1,-1, }, // ..ooo..
      {  5, 6, 7,16, 8, 9,10, }, // oooxooo
      { -1,-1,11,12,13,-1,-1, }, // ..ooo..
      { -1,-1,-1,14,-1,-1,-1, }, // ...o...
      { -1,-1,-1,15,-1,-1,-1, }, // ...o...
    },
  },
  { // (10) 7x7cross + 3x3 square
    7, 9, 3, 3,
    {
      { -1,-1,-1, 0,-1,-1,-1, }, // ...o...
      { -1,-1,-1, 1,-1,-1,-1, }, // ...o...
      { -1,-1, 2, 3, 4,-1,-1, }, // ..ooo..
      {  5, 6, 7, 8, 7, 6, 5, }, // ooox+++
      { -1,-1, 4, 3, 2,-1,-1, }, // ..+++..
      { -1,-1,-1, 1,-1,-1,-1, }, // ...+...
      { -1,-1,-1, 0,-1,-1,-1, }, // ...+...
    },
  },
};


BLFParam::BLFParam()
{
  pattern = BLF_DEF_PATTERN_IDX; 
  len = blfPattern[pattern].numCoef-1;  // center coef is not coded
  offset[0] = 0;
  offset[1] = 0;
  offset[2] = 0;
  coef[0] = NULL;
  coef[1] = NULL;
  coef[2] = NULL;
  onFlag = NULL;
  enabled = false;
}

BLFParam::BLFParam( BLFParam& orig )
{
  *this = orig;
}

BLFParam::~BLFParam()
{
  if (coef[0]) { delete[] coef[0]; }
  if (coef[1]) { delete[] coef[1]; }
  if (coef[2]) { delete[] coef[2]; }
  if (onFlag)  { delete[] onFlag;  }
}

BLFParam& BLFParam::operator=( BLFParam& orig )
{
  if (this == &orig) return *this;

  pattern = orig.pattern;
  len = orig.len;

  offset[0] = orig.offset[0];
  offset[1] = orig.offset[1];
  offset[2] = orig.offset[2];
  if (coef[0]) { delete[] coef[0]; coef[0] = NULL; }
  if (coef[1]) { delete[] coef[1]; coef[1] = NULL; }
  if (coef[2]) { delete[] coef[2]; coef[2] = NULL; }
  if (!coef[0]) { coef[0] = new Int[len]; }
  if (!coef[1]) { coef[1] = new Int[len]; }
  if (!coef[2]) { coef[2] = new Int[len]; }
  for (Int i=0; i<len; i++)
  {
    if (orig.coef[0]) coef[0][i] = orig.coef[0][i];
    if (orig.coef[1]) coef[1][i] = orig.coef[1][i];
    if (orig.coef[2]) coef[2][i] = orig.coef[2][i];
  }

  if (onFlag) { delete[] onFlag; onFlag = NULL; }
  if (!onFlag) { onFlag = new Int[BL_FILTSW_NBLKS]; }
  if (orig.onFlag)
  {
    for (Int i=0; i<BL_FILTSW_NBLKS; i++)
    {
      onFlag[i] = orig.onFlag[i];
    }
  }
  enabled = orig.enabled;
  return *this;
}



const BLFPattern* TComBaseLayerFilter::getCurrentBLFPattern()
{
  return &blfPattern[BLF_DEF_PATTERN_IDX];
}

BLFParam* TComBaseLayerFilter::initBLFParam( Int pattern )
{
  if (pattern < 0)
  {
    pattern = BLF_DEF_PATTERN_IDX;
  }
  m_blfParam.pattern = pattern;
  m_blfParam.len = blfPattern[pattern].numCoef -1;

  m_blfParam.offset[0] = 0;
  m_blfParam.offset[1] = 0;
  m_blfParam.offset[2] = 0;
  if (!m_blfParam.coef[0]) { m_blfParam.coef[0] = new Int[m_blfParam.len]; }
  if (!m_blfParam.coef[1]) { m_blfParam.coef[1] = new Int[m_blfParam.len]; }
  if (!m_blfParam.coef[2]) { m_blfParam.coef[2] = new Int[m_blfParam.len]; }
  for (Int i=0; i<m_blfParam.len; i++)
  {
    m_blfParam.coef[0][i] = 0;
    m_blfParam.coef[1][i] = 0;
    m_blfParam.coef[2][i] = 0;
  }

  if (!m_blfParam.onFlag) { m_blfParam.onFlag = new Int[BL_FILTSW_NBLKS]; }
  for (Int i=0; i<BL_FILTSW_NBLKS; i++)
  {
    m_blfParam.onFlag[i] = 1;
  }
  m_blfParam.enabled = false;

  return &m_blfParam;
}

void TComBaseLayerFilter::findWienerFilterCoef( TComPicYuv* org, TComPicYuv* deg, BLFParam* blfp )
{
  findWienerFilterCoef( org, deg, blfp, 0 );
  findWienerFilterCoef( org, deg, blfp, 1 );
  findWienerFilterCoef( org, deg, blfp, 2 );
}

void TComBaseLayerFilter::findWienerFilterCoef( TComPicYuv* org, TComPicYuv* deg, BLFParam* blfp, Int chromaIdc )
{
  const BLFPattern &wf = blfPattern[blfp->pattern];
  const Int SIZE = wf.size;
  const Int CENTER_IDX = (wf.center_y*SIZE+wf.center_x);
  const Int NUM_EQU = wf.numCoef;                   // number of different coefficients
  const Int MAX_N = BLF_NUM_COEF_MAX+1;
  double v[MAX_N] = {0.0};
  double m[MAX_N][MAX_N] = {{0.0}};
  Int& offset = blfp->offset[chromaIdc];
  Int* packedCoeff = blfp->coef[chromaIdc];

  // gather pixel values
  Int orgStride = chromaIdc? org->getCStride(): org->getStride();
  Int degStride = chromaIdc? deg->getCStride(): deg->getStride();
  Pel* pOrg = (chromaIdc==2)? org->getCrAddr(): (chromaIdc==1)? org->getCbAddr(): org->getLumaAddr();
  Pel* pDeg = (chromaIdc==2)? deg->getCrAddr(): (chromaIdc==1)? deg->getCbAddr(): deg->getLumaAddr();
  const Int picw  = org->getWidth();
  const Int pich  = org->getHeight();

  // for all pixels
  for (Int by = 0; by < BL_FILTSW_NBLK_VERT; by++ )
  {
    for (Int bx = 0; bx < BL_FILTSW_NBLK_HORZ; bx++ )
    {
      const Int bidx = by*BL_FILTSW_NBLK_HORZ+bx;
      if (!blfp->onFlag[bidx]) continue;
      Int x0,y0,x1,y1;
      xGetBlockRect(x0, y0, x1, y1, bx, by, picw, pich, chromaIdc);

      for (Int py = y0; py < y1; py++)
      {
        Int py0 = py-wf.center_y;
        for (Int px = x0; px < x1; px++)
        {
          Int px0 = px-wf.center_x;
          double orgPix = pOrg[py*orgStride + px];

          // for all equations of coeff
          double degPixSum[MAX_N-1] = {};
          for (Int ey = 0; ey < SIZE; ey++) 
          {
            for (Int ex = 0; ex < SIZE; ex++) 
            {
              //if ( wf.coefIdx[ey][ex] < 0) continue;  //  for non-square filter
              degPixSum[wf.coefIdx[ey][ex]] += pDeg[(py0+ey)*degStride + (px0+ex)];
            }
          }

          for (Int eq = 0; eq < NUM_EQU; eq++ )
          {
            v[eq] += orgPix * degPixSum[eq];
            for (Int eq1 = 0; eq1 <= eq; eq1++ )
            {
              m[eq][eq1] += degPixSum[eq] * degPixSum[eq1];
            }
          }
        }
      }
    }
  }
  
  for (Int eq = 0; eq < NUM_EQU; eq++ )
  {
    for (Int eq1 = eq+1; eq1 < NUM_EQU; eq1++ )
    {
      m[eq][eq1] = m[eq1][eq];
    }
  }


  // calculate coefficients

  int i, j, k;
  double alpha[MAX_N][MAX_N] = {{0.0}};
  double beta [MAX_N][MAX_N] = {{0.0}};
  double coeffDbl[MAX_N] = {0.0};

  Int N = NUM_EQU;        // compute required number of taps + filter DC offset
  Bool filterFlag = true;


  // add ridge parameter
  double reg = 0;

  reg = m[0][0]*1E-5; 

  for (i = 0; i < N; i++)
  {
    m[i][i] += reg;
  }

  // Step 1: LU Decomposition

  //initialize the ii-components
  for(i = 0; i < N; i++)
  {
    alpha[i][i] = 1.0;
  }
  for(j = 0; j < N; j++)
  {
    for(i = 0; i <=j; i++)
    {
      beta[i][j] = m[i][j];
      for(k = 0; k < i; k++)
      {
        beta[i][j] -= alpha[i][k]*beta[k][j];
      }
    }
    if(fabs(beta[j][j])<0.000001)
    {
      filterFlag = false;
    }
    else
    {
      for(i = j+1; i < N; i++)
      {
        alpha[i][j] = m[i][j];
        for(k = 0; k < j; k++)
        {
          alpha[i][j] -= alpha[i][k]*beta[k][j];
        }
        alpha[i][j] = alpha[i][j]/beta[j][j];
      }
    }
  }

  if(filterFlag)
  {
    double y[MAX_N] = {0.0};
    // Step 2: Solving L*y = b, forward substitution
    y[0]=v[0];
    for(i = 1; i < N; i++)
    {
      y[i] = v[i];
      for(j = 0; j < i; j++)
      {
        y[i] -= alpha[i][j]*y[j];
      }
    }
    // Step 3: Solving U*x = y, back substitution
    coeffDbl[N-1] = y[N-1]/beta[N-1][N-1];
    for(i = N-2; i >= 0; i--)
    {
      coeffDbl[i] = y[i];
      for(j = i+1; j < N; j++)
      {
        coeffDbl[i] -= beta[i][j]*coeffDbl[j];
      }
      coeffDbl[i] = coeffDbl[i]/beta[i][i];
    }

//    xPrintCoef( wf, coeffDbl );
  }
  else
  {
    coeffDbl[CENTER_IDX] = 1.0;
  }

  // quantize coefficients and adjust sum of them
  Int coeff[MAX_N] = {0};
  Int sum=0;
  double adj = 0;
  offset =0;
  sum = xQuantCoef(wf, coeffDbl, adj, coeff);
  const Int nStep = 20;
  const double step = (0.5/BLF_COEF_ACC/nStep);
  if (sum < BLF_COEF_ACC-1) {
    for (adj = step; adj < step*nStep; adj += step)
    {
      sum = xQuantCoef(wf, coeffDbl, -adj, coeff);
      if (BLF_COEF_ACC-1 <= sum) break;
    }
  }
  else if (BLF_COEF_ACC+1 < sum) {
    for (adj = step; adj < step*nStep; adj += step)
    {
      sum = xQuantCoef(wf, coeffDbl, adj, coeff);
      if (sum <= BLF_COEF_ACC+1) break;
    }
  }
   
  xPackCoefArray( wf, coeff, packedCoeff );
}


void TComBaseLayerFilter::convolute( TComPicYuv* src, TComPicYuv* dst, BLFParam* blfp )
{
  convolute( src, dst, blfp, 0 );
  convolute( src, dst, blfp, 1 );
  convolute( src, dst, blfp, 2 );
}

// TODO: omit calculation on unnecessary blocks
void TComBaseLayerFilter::convolute( TComPicYuv* src, TComPicYuv* dst, BLFParam *blfp, Int chromaIdc )
{
  // filtering
  if (!src || !dst || !blfp) return;

  const BLFPattern &wf = blfPattern[blfp->pattern];
  const Int SIZE = wf.size;
  Int& offset = blfp->offset[chromaIdc];
  Int* packedCoeff = blfp->coef[chromaIdc];
  Int coeff[BLF_NUM_COEF_MAX] = {0};

  Int srcStride = chromaIdc? src->getCStride(): src->getStride();
  Int dstStride = chromaIdc? dst->getCStride(): dst->getStride();
  Pel* pSrc = (chromaIdc==2)? src->getCrAddr(): (chromaIdc==1)? src->getCbAddr(): src->getLumaAddr();
  Pel* pDst = (chromaIdc==2)? dst->getCrAddr(): (chromaIdc==1)? dst->getCbAddr(): dst->getLumaAddr();
  Int width  = src->getWidth()  >> (chromaIdc? 1:0);
  Int height = src->getHeight() >> (chromaIdc? 1:0);

  xUnpackCoefArray( wf, packedCoeff, coeff );
#if 1
  for (Int py = 0; py < height; py++)
  {
    for (Int px = 0; px < width; px++)
    {
      // convolution
      Int value = 0;
      for (Int fy = 0; fy < SIZE; fy++) 
      {
        for (Int fx = 0; fx < SIZE; fx++) 
        {
          Int cy = py+fy-wf.center_y;
          Int cx = px+fx-wf.center_x;
          Int fc = fy*SIZE+fx; // coeff idx
          value += coeff[fc] * pSrc[cy*srcStride + cx];
        }
      }
      pDst[py*dstStride+px] = Clip( (value + offset + (BLF_COEF_ACC/2))>>BLF_COEF_ACC_BITS );
    }
  }  
#else
  // avoid unnecessary filtering (NOT finished: must modify other associated codes)
  for (Int by = 0; by < BL_FILTSW_NBLK_VERT; by++ )
  {
    for (Int bx = 0; bx < BL_FILTSW_NBLK_HORZ; bx++ )
    {
      const Int bidx = by*BL_FILTSW_NBLK_HORZ+bx;
      if (!blfp->onFlag[bidx]) continue;
      Int x0,y0,x1,y1;
      xGetBlockRect(x0, y0, x1, y1, bx, by, width, height, chromaIdc);

      //printf("convo[%d]: (%d,%d - %d,%d)  flag=%d\n",chromaIdc, x0,y0,x1,y1, blfp->onFlag[bidx]);
      if (!blfp->onFlag[bidx]) continue;

      for (Int py = y0; py < y1; py++)
      {
        for (Int px = x0; px < x1; px++)
        {
          // convolution
          Int value = 0;
          for (Int fy = 0; fy < SIZE; fy++) 
          {
            for (Int fx = 0; fx < SIZE; fx++) 
            {
              Int cy = py+fy-wf.center_y;
              Int cx = px+fx-wf.center_x;
              Int fc = fy*SIZE+fx; // coeff idx
              value += coeff[fc] * pSrc[cy*srcStride + cx];
            }
          }
          pDst[py*dstStride+px] = Clip( (value + offset + (BLF_COEF_ACC/2))>>BLF_COEF_ACC_BITS );
        }
      }  
    }
  }
#endif
}


void TComBaseLayerFilter::calcSAD(TComPicYuv*src1,TComPicYuv*src2, UInt x0, UInt y0, UInt w, UInt h, UInt64& retDist)
{
  UInt stride1 = src1->getStride();
  UInt stride2 = src2->getStride();
  UInt offset1 = y0*stride1 + x0;
  UInt offset2 = y0*stride2 + x0;
  Pel* p1 = src1->getLumaAddr() + offset1;
  Pel* p2 = src2->getLumaAddr() + offset2;

  retDist=0;

  for(Int y = 0; y < h; y++ )
  {
    for(Int x = 0; x < w; x++ )
    {
      Int diff = p1[x] - p2[x];
      retDist += (diff<0)? -diff: diff; //SAD
    }
    p1 += stride1;
    p2 += stride2;
  }
  stride1 = src1->getCStride();
  stride2 = src2->getCStride();
  offset1 = (y0>>1)*stride1 + (x0>>1);
  offset2 = (y0>>1)*stride2 + (x0>>1);
  for (Int c=0;c<=1;c++)
  {
    p1 = ( c? src1->getCrAddr(): src1->getCbAddr() ) + offset1;
    p2 = ( c? src2->getCrAddr(): src2->getCbAddr() ) + offset2;
    for(Int y = 0; y < h/2; y++ )
    {
      for(Int x = 0; x < w/2; x++ )
      {
        Int diff = p1[x] - p2[x];
        retDist += (diff<0)? -diff: diff; //SAD
      }
      p1 += stride1;
      p2 += stride2;
    }
  }

}


// select better block and composite picture
void TComBaseLayerFilter::decideBetterPicOnBlockBasis(TComPicYuv* org, TComPicYuv** srcs, Int numSrc, BLFParam* blfp )
{
  decideBetterPicOnBlockBasis(org, srcs, numSrc, blfp->onFlag );
  blfp->enabled = false;

  for (Int i=0; i<BL_FILTSW_NBLKS; i++)
  {
    if (blfp->onFlag[i] == 1)
    {
      blfp->enabled = true;
      break;
    }
  }
}

void TComBaseLayerFilter::decideBetterPicOnBlockBasis(TComPicYuv* org, TComPicYuv** srcs, Int numSrc, Int *idxBest )
{
  const Int picw = org->getWidth();
  const Int pich = org->getHeight();

  memset(idxBest, 0, sizeof(Int)*BL_FILTSW_NBLKS);

  for (Int by = 0; by < BL_FILTSW_NBLK_VERT; by++ )
  {
    for (Int bx = 0; bx < BL_FILTSW_NBLK_HORZ; bx++ )
    {
      const Int bidx = by*BL_FILTSW_NBLK_HORZ+bx;
      Int x0,y0,x1,y1;
      xGetBlockRect(x0, y0, x1, y1, bx, by, picw, pich, 0);

      Int w = x1-x0;
      Int h = y1-y0;
      UInt64 distBest = ~(UInt64)0;

      for (Int i=0; i<numSrc; i++)
      {
        UInt64 dist = 0;
        calcSAD( org, srcs[i], x0, y0, w, h, dist );

        if (dist < distBest)
        {
          distBest = dist;
          idxBest[bidx] = i;
        }
      }
    }
  }
}


void TComBaseLayerFilter::compositeBetterPicOnBlockBasis(TComPicYuv** srcs, TComPicYuv* dst, Int* idxBest )
{
  const Int picw = dst->getWidth();
  const Int pich = dst->getHeight();

  for (Int by = 0; by < BL_FILTSW_NBLK_VERT; by++ )
  {
    for (Int bx = 0; bx < BL_FILTSW_NBLK_HORZ; bx++ )
    {
      const Int bidx = by*BL_FILTSW_NBLK_HORZ+bx;
      Int x0,y0,x1,y1;
      xGetBlockRect(x0, y0, x1, y1, bx, by, picw, pich, 0);

      // copy a block
      TComPicYuv* srcBest = srcs[idxBest[bidx]];
      srcBest->copyRegionToRegionYuv( dst, x0, y0, x1-x0, y1-y0 );
    }
  }
}

void TComBaseLayerFilter::compressChromaCoef( BLFParam* blfp )
{
  for (Int i=0; i<blfp->len; i++)
  {
    blfp->coef[2][i] -= blfp->coef[1][i]; 
  }
}

void TComBaseLayerFilter::decompressChromaCoef( BLFParam* blfp )
{
  for (Int i=0; i<blfp->len; i++)
  {
    blfp->coef[2][i] += blfp->coef[1][i]; 
  }
}

//
// protected
//

// pack coeff array
void TComBaseLayerFilter::xPackCoefArray( const BLFPattern& wf, Int* coeff, Int* packedCoeff )
{
  //Int pidx = 0;
  for (Int i = 0; i < wf.size; i++)
  {
    for (Int j = 0; j < wf.size; j++)
    {
      Int ci = i*wf.size+j;
      Int pci = wf.coefIdx[i][j];
      if ( 0 <= pci && pci < wf.numCoef-1 )
      {
        packedCoeff[pci] = coeff[ci];
      }
      if (pci == wf.numCoef-2) return;
    }
  }
}

// expand packed coeff array
void TComBaseLayerFilter::xUnpackCoefArray( const BLFPattern& wf, Int* packedCoeff, Int* coeff )
{
  Int sum = 0;
  for (Int i = 0; i < wf.size; i++)
  {
    for (Int j = 0; j < wf.size; j++)
    {
      Int ci = i*wf.size+j;
      Int pci = wf.coefIdx[i][j];
      coeff[ci] = (0 <= pci && pci < wf.numCoef-1 )? packedCoeff[pci]: 0;
      sum += coeff[ci];
    }
  }
  const Int CENTER_IDX = wf.center_y*wf.size+wf.center_x;
  coeff[ CENTER_IDX ] = BLF_COEF_ACC - sum;

}

// quantize coefficient array of double and return sum
Int TComBaseLayerFilter::xQuantCoef( const BLFPattern& wf, double *coeffDbl, double adj, Int *coeff )
{
  Int sum = 0;
  for (Int i = 0; i < wf.size; i++)
  {
    for (Int j = 0; j < wf.size; j++)
    {
      Int pci = wf.coefIdx[i][j];
      if ( pci < 0 ) continue;
      Int ci = i*wf.size+j;

      double r = ((coeffDbl[pci]-adj)*BLF_COEF_ACC > 0)? 0.5: -0.5;
      coeff[ci] = floor((coeffDbl[pci]-adj)*BLF_COEF_ACC + r);

      sum += coeff[ci];
    }
  }
  return sum;
}

Void TComBaseLayerFilter::xGetBlockRect(Int& x0, Int& y0, Int& x1, Int& y1, Int bx, Int by, Int picw, Int pich, Int chromaIdc)
{
  picw >>= chromaIdc? 1: 0;
  pich >>= chromaIdc? 1: 0;
  const Int blkw = picw / BL_FILTSW_NBLK_HORZ;
  const Int blkh = pich / BL_FILTSW_NBLK_VERT;
  Int bits = (chromaIdc==0)? 6 : 5;
  Int e = (1<<(bits-1)) -1;

  y0 = ((blkh* by    + e) >> bits ) << bits;
  y1 = ((blkh*(by+1) + e) >> bits ) << bits;
  x0 = ((blkw* bx    + e) >> bits ) << bits;
  x1 = ((blkw*(bx+1) + e) >> bits ) << bits;
  if ( by == BL_FILTSW_NBLK_VERT-1 ) { y1 = pich; }
  if ( bx == BL_FILTSW_NBLK_HORZ-1 ) { x1 = picw; }
}

// debug output
void TComBaseLayerFilter::xPrintCoef( const BLFPattern& wf, Int *coeff, Int offset )
{
  Int sum = 0;
  printf("# TComBaseLayerFilter: coeff\n");
  for (Int i = 0; i < wf.size; i++)
  {
    printf("# ");
    for (Int j = 0; j < wf.size; j++)
    {
      Int c = coeff[i*wf.size+j];
      printf(" %5d", c);
      sum += c;
    }
    printf("\n");
  }
  printf("# sum=%d\n", sum);
}

void TComBaseLayerFilter::xPrintCoef( const BLFPattern& wf, double *coeffDbl )
{
  double sumDbl=0;
  printf("# TComBaseLayerFilter: coeffDbl\n");
  for (Int i = 0; i < wf.size; i++)
  {
    printf("# ");
    for (Int j = 0; j < wf.size; j++)
    {
      Int idx = wf.coefIdx[i][j];
      if ( idx < 0 ) {
        printf("      ");
      }
      else
      {
        printf(" %5.2f", coeffDbl[idx]);
        sumDbl += coeffDbl[idx];
      }
    }
    printf("\n");
  }
  printf("# sum=%5.2f\n", sumDbl);
}

void TComBaseLayerFilter::printFilterFlags( Int *flags )
{
  printf("# baseLayerFilterFlags=\n");
  for (Int j = 0; j < BL_FILTSW_NBLK_VERT; j++)
  {
    printf("# ");
    for (Int i = 0; i < BL_FILTSW_NBLK_HORZ; i++)
    {
      printf(" %1d", flags[j*BL_FILTSW_NBLK_HORZ+i]);
    }
    printf("\n");
  }
  printf("\n");
}

